import inspect
import json
import functools


__all__ = ['Serializer', 'register', 'unregister', 'get_serializer',
           'base_create_object', 'RegisterMetaclass',
           'dumps', 'dump', 'loads', 'load', 'default', 'object_hook']


def base_create_object(cls):
    """Primitive create object. This will work if parent class overrides __new__ with *args, **kwargs."""
    # Create the object without failing from __init__ args (if parent class overrides __new__ with *args, **kwargs)
    obj = cls.__new__(cls)

    # Try to initialize properly (may fail because of no args)
    try:
        obj.__init__()
    except (TypeError, Exception):
        pass

    return obj


# ========== Serializers ==========
SERIALIZERS = []  # Serializer
SERIALIZER_TYPE = 'SERIALIZER_TYPE'
SERIALIZER_OBJ = 'SERIALIZER_OBJ'


class Serializer(object):
    """Serializer that will try to use an objects "__getstate__" and "__setstate__" methods.

    Encode functions should take in an object and return a dictionary

    Decode functions should take in a dictionary and return a new object from that objects data.
    """
    def __init__(self, cls, encode=None, decode=None):
        self.cls = cls
        self.serializer_name = getattr(cls, '__qualname__', '{}.{}'.format(cls.__module__, cls.__name__))

        if encode is not None:
            self.encode = encode
        if decode is not None:
            self.decode = decode

    def encode(self, obj):
        return obj.__getstate__()

    def decode(self, obj):
        try:
            new_obj = self.cls()
        except Exception as err:
            try:
                new_obj = base_create_object(self.cls)
            except (ValueError, TypeError, Exception):
                raise err

        new_obj.__setstate__(obj)
        return new_obj


def register(cls_obj=None, encode=None, decode=None):
    """Register a serializer class.

    By default the serializer will use the class type "__getstate__" and "__setstate__" methods if an encode and decode
    method is not given.

    Args:
        cls_obj (class/type): Class/Type/object to register the serializer encode and decode methods with.
        encode (function): Function to convert an object of this type to a dictionary.
        decode (function): Function to convert a dictionary representing this type into an object.

    Returns:
        cls (class/type/function): Class/Type that was registered OR decorator function.
    """
    global SERIALIZERS

    # ===== As Decorator =====
    if cls_obj is None:
        def wrapper(cls):
            register(cls_obj=cls_obj, encode=encode, decode=decode)
            return cls
        return wrapper

    # ===== Register the class as a serializer =====
    try:
        if not inspect.isclass(cls_obj):
            cls_obj = cls_obj.__class__
    except (AttributeError, Exception):
        pass

    # Save the serializer class
    serializer = Serializer(cls=cls_obj, encode=encode, decode=decode)
    registered = False
    for i, ser in enumerate(SERIALIZERS):
        if ser.cls == cls_obj:
            SERIALIZERS[i] = serializer
            registered = True

    if not registered:
        SERIALIZERS.append(serializer)

    return cls_obj


def unregister(cls_obj):
    """Remove a registered serializer."""
    global SERIALIZERS
    try:
        if not inspect.isclass(cls_obj):
            cls_obj = cls_obj.__class__
    except (AttributeError, Exception):
        pass

    for i in range(len(SERIALIZERS)):
        ser = SERIALIZERS[i]
        if ser.cls == cls_obj:
            SERIALIZERS.pop(i)
            break


def get_serializer(cls_obj):
    """Return a serializer class for the given type."""
    global SERIALIZERS

    # ===== Serializer from class/type =====
    try:
        if not isinstance(cls_obj, str) and not inspect.isclass(cls_obj):
            cls_obj = cls_obj.__class__
    except (AttributeError, Exception):
        pass

    first_sub = None
    for ser in SERIALIZERS:
        try:
            # Check if given object is class or matching string serializer_name
            if ser.cls == cls_obj or ser.serializer_name == cls_obj:
                return ser
            elif first_sub is None and issubclass(cls_obj, ser.cls):
                first_sub = ser
        except (TypeError, ValueError, Exception):
            pass

    return first_sub


# ========== Default Message Object ==========
class RegisterMetaclass(type):
    """Metaclass that automatically registers subclasses """
    def __new__(mcs, name, bases, class_dict):
        cls = type.__new__(mcs, name, bases, class_dict)
        register(cls)
        return cls


# ========== JSON Parsers ==========
_default_encoder = json._default_encoder
_default_decoder = json._default_decoder


def default(obj):
    """Default function for how to serialize an object."""
    # Get serializer
    ser = get_serializer(obj)

    if ser is not None:
        # Get the state dictionary
        d = ser.encode(obj)
        if not isinstance(d, dict):
            d = {SERIALIZER_OBJ: d}
        d[SERIALIZER_TYPE] = ser.serializer_name
        return d

    return _default_encoder.default(obj)


def object_hook(obj):
    """Default function for how to deserialize an object."""
    if isinstance(obj, dict) and SERIALIZER_TYPE in obj:
        name = obj.pop(SERIALIZER_TYPE, None)
        obj = obj.pop(SERIALIZER_OBJ, obj)
        if name is not None:
            ser = get_serializer(name)
            if ser is not None:
                return ser.decode(obj)

    if _default_decoder.object_hook is not None:
        return _default_decoder.object_hook(obj)
    return obj


@functools.wraps(json.dumps)
def dumps(obj, **kwargs):
    kwargs['default'] = default
    return json.dumps(obj, **kwargs)


@functools.wraps(json.dump)
def dump(obj, fp, **kwargs):
    kwargs['default'] = default
    return json.dump(obj, fp, **kwargs)


@functools.wraps(json.loads)
def loads(s, **kwargs):
    kwargs['object_hook'] = object_hook
    return json.loads(s, **kwargs)


@functools.wraps(json.load)
def load(fp, **kwargs):
    kwargs['object_hook'] = object_hook
    return json.load(fp, **kwargs)



import ast
import numpy as np
from serial_json.interface import register


__all__ = ['np_to_dict', 'np_from_dict', 'rec_from_dict']


def np_to_dict(obj):
    """Convert a numpy array to a json serializable dictionary."""
    dtype = str(obj.dtype)
    if getattr(obj.dtype, 'names', None) is not None:
        dtype = str(obj.dtype.descr)  # Record array and structured array support
    return {'data': bytes(obj.data),  # REQUIRES bytes_support!
            'shape': ','.join((str(i) for i in obj.shape)), 'dtype': dtype}


def np_from_dict(obj):
    """Convert a json dictionary to a numpy array."""
    byts = obj.get('data', b'')
    shape = tuple(int(i) for i in obj.get('shape', '-1').split(','))
    dtype = obj.get('dtype', '<f4')
    try:
        dtype = ast.literal_eval(dtype)
    except (ValueError, TypeError, Exception):
        pass
    dtype = np.dtype(dtype)

    # Create an array and override the memoryview
    arr = np.empty(shape, dtype=dtype)
    arr.data = byts
    return arr


def rec_from_dict(obj):
    """Convert a json dictionary to a numpy record array."""
    return np_from_dict(obj).view(np.recarray)


register(np.ndarray, np_to_dict, np_from_dict)
register(np.recarray, np_to_dict, rec_from_dict)  # Record array support



import copy
from inspect import signature, Signature, Parameter
from typing import Any, Callable
from collections import OrderedDict

from .interface import register


__all__ = ['MISSING', 'field', 'field_property', 'DataclassMeta', 'DataClass', 'dataclass', 'Message']


class MISSING(object):
    pass


class field(object):
    """Dataclass field attribute

    Args:
        default (object/Any): Default value on init.
        default_factory (function/Callable): Function that returns the default value.
        repr (bool)[True]: Include this field in the repr.
        hash (bool)[None]: If True include in __hash__. If None include in __hash__ if compare is True.
        init (bool)[True]: Include this field in the __init__. If False set the value from the default value.
        compare (bool)[True]: Include this field in the __eq__ comparison.
        metadata (dict)[None]: Unused.
        dict (bool)[True]: Include this field in the dict() method.
        skip_dict (object/Any)[MISSING]: Do not include this field in the dict() if the set value equals this value.
        skip_repr (object/Any)[MISSING]: Do not include this field in the __repr__ if the set value equals this value.
        name (str)[MISSING]: Field variable name. This is automatically set.
        type (object/type): Field type. Used in __init__ doc string and annotation.
    """
    def __init__(self, default=MISSING, default_factory=MISSING, repr=True, hash=None, init=True, compare=True,
                 metadata=None, dict=True, skip_dict=MISSING, skip_repr=MISSING, name=MISSING, type=Any, **kwargs):
        super().__init__()
        self.default = default
        self.default_factory = default_factory
        self.repr = repr
        self.skip_repr = skip_repr
        self.hash = hash
        self.init = init
        self.compare = compare
        self.metadata = metadata
        self.dict = dict
        self.skip_dict = skip_dict
        self.type = type
        self.name = name

        # Save other given attributes
        for k, v in kwargs.items():
            setattr(self, k, v)

    def set_name(self, n, replace=True):
        """Set the name."""
        if self.name == MISSING or replace:
            self.name = n

    def is_required(self):
        """Return if this argument is required in init."""
        return self.init and self.default == MISSING and self.default_factory == MISSING

    def set_default(self, default):
        """Set the default value."""
        self.default = default
        return self

    def set_default_factory(self, default_factory):
        """Set the default factory function to get the default value."""
        self.default_factory = default_factory
        return self

    def get_default_value(self, obj=None):
        """Return this object's field default value.

        Args:
            obj (object)[None]: Instance object for this field.

        Raises:
            error (TypeError): If a default is not set and this argument is required.

        Returns:
            value (object/Any): Default Value.
        """
        if self.default != MISSING:
            try:
                return copy.copy(self.default)
            except (TypeError, ValueError, Exception):
                pass
            return self.default
        elif self.default_factory != MISSING:
            try:
                return self.default_factory.__get__(obj, type(obj))()
            except (TypeError, ValueError, Exception):
                pass
            return self.default_factory()
        raise TypeError('missing required argument: {}'.format(self.name))

    def get_default_str(self):
        """Return this fields default value for the doc string."""
        if self.default != MISSING:
            try:
                return copy.copy(self.default)
            except (TypeError, ValueError, Exception):
                pass
            return self.default
        elif self.default_factory != MISSING:
            return str(self.default_factory)
        return None

    def get_type(self):
        """Return this field's type as a string name."""
        typ = Any
        if self.type is not None:
            typ = self.type
        elif self.default != MISSING:
            typ = type(self.default)
        elif self.default_factory != MISSING:
            try:
                typ = type(self.default_factory())
            except (TypeError, ValueError, AttributeError, Exception):
                pass

        return typ

    def get_type_str(self):
        typ = self.get_type()
        try:
            return typ.__name__
        except (AttributeError, Exception):
            return typ

    def as_parameter(self):
        kind = Parameter.POSITIONAL_ONLY
        default = Parameter.empty
        annotation = self.get_type()
        if not self.is_required():
            kind = Parameter.POSITIONAL_OR_KEYWORD
            default = self.get_default_str()

        return Parameter(name=self.name, kind=kind, default=default, annotation=annotation)


class field_property(field):
    """Field that works as a property."""
    def __init__(self, fget=None, fset=None, fdel=None,
                 default=MISSING, default_factory=MISSING, repr=True, hash=None, init=True, compare=True,
                 metadata=None, dict=True, skip_dict=MISSING, skip_repr=MISSING, name=MISSING, type=None, **kwargs):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        super().__init__(default=default, default_factory=default_factory, repr=repr, hash=hash, init=init,
                         compare=compare, metadata=metadata, dict=dict, skip_dict=skip_dict, skip_repr=skip_repr,
                         name=name, type=type, **kwargs)

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        return self.fget(obj)

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        self.fdel(obj)

    def getter(self, fget: Callable[[Any], Any]) -> 'field_property':
        self.fget = fget
        return self

    __call__ = getter

    def setter(self, fset: Callable[[Any, Any], None]) -> 'field_property':
        self.fset = fset
        return self

    def deleter(self, fdel: Callable[[Any], None]) -> 'field_property':
        self.fdel = fdel
        return self


class DataclassMeta(type):
    def __new__(cls, name, bases, dct):
        new_cls = super().__new__(cls, name, bases, dct)
        return cls.dataclass(new_cls)

    @classmethod
    def dataclass(cls, new_cls=None, init=True, repr=True, eq=True, order=False, unsafe_hash=False, dict=True,
                       frozen=False, **kwargs):
        """Return the given class as a dataclass."""
        if new_cls is None:
            def decorator(new_cls):
                return cls.dataclass(new_cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                                          dict=dict, frozen=frozen, **kwargs)
            return decorator

        cls.make_fields(new_cls)

        if init and new_cls.__init__ == object.__init__ and not getattr(new_cls.__init__, '__is_dataclass__', False):
            # Set the __init__ function
            def __init__(self, *args, **kwargs):
                return cls.init_func(self, *args, **kwargs)
            __init__.__is_dataclass__ = True
            new_cls.__init__ = __init__

        if init and getattr(new_cls.__init__, '__is_dataclass__', False):
            # Make the signature and doc string for the __init__ function
            try:
                new_cls.__init__.__signature__ = cls.make_init_signature(new_cls)
                doc = cls.make_init_docs(new_cls)
                if not new_cls.__doc__:
                    new_cls.__doc__ = doc
                new_cls.__init__.__doc__ = doc
            except AttributeError:
                pass

        if dict and 'dict' not in new_cls.__dict__:
            new_cls.dict = cls.asdict
        if dict and not hasattr(new_cls, '__getstate__'):
            new_cls.__getstate__ = cls.getstate_func
        if dict and not hasattr(new_cls, '__setstate__'):
            new_cls.__setstate__ = cls.setstate_func

        if new_cls.__hash__ == object.__hash__:
            new_cls.__hash__ = cls.hash_func

        if eq and new_cls.__eq__ == object.__eq__:
            new_cls.__eq__ = cls.compare_func

        if repr and new_cls.__repr__ == object.__repr__:
            new_cls.__repr__ = cls.repr_func

        new_cls.__is_frozen__ = frozen

        # Register for serialization
        register(new_cls)

        return new_cls

    @staticmethod
    def make_fields(cls):
        bases = cls.__bases__

        # Setup annotations and fields
        if not hasattr(cls, '__annotations__'):
            cls.__annotations__ = OrderedDict()

        if not hasattr(cls, '__fields__'):
            cls.__fields__ = OrderedDict()
        elif any(getattr(base, '__fields__', {}) is cls.__fields__ for base in bases):
            cls.__fields__ = cls.__fields__.copy()

        if not hasattr(cls, '__dataclass_properties__'):
            cls.__dataclass_properties__ = {'allow_extra_init': True}
        elif any(getattr(base, '__dataclass_properties__', {}) is cls.__dataclass_properties__ for base in bases):
            cls.__dataclass_properties__ = cls.__dataclass_properties__.copy()

        # Make annotations
        for name, attr in cls.__dict__.items():
            if name.startswith('__'):
                continue

            if name in cls.__fields__ and name not in cls.__annotations__:
                cls.__annotations__[name] = cls.__fields__[name].get_type()
            elif isinstance(attr, field) and name not in cls.__annotations__:
                cls.__annotations__[name] = attr.get_type()
            elif isinstance(attr, property) and name not in cls.__annotations__:
                ret_typ = signature(attr.fget).return_annotation
                if ret_typ == Signature.empty:
                    ret_typ = Any
                cls.__annotations__[name] = ret_typ

        # Make fields
        for name, typ in cls.__annotations__.items():
            if name not in cls.__fields__:
                f = cls.__dict__.get(name, None)
                if not isinstance(f, field):
                    if name in cls.__dict__:
                        f = cls.__fields__.get(name, field(default=cls.__dict__[name], type=typ))
                    else:
                        f = cls.__fields__.get(name, field(default=MISSING, type=typ))
                else:
                    f.type = typ
                cls.__fields__[name] = f
                f.set_name(name, replace=False)

    @staticmethod
    def make_init_docs(cls):
        def args():
            for f in cls.__fields__.values():
                if f.init:
                    if f.is_required():
                        yield '{name} ({type}): {name} value\n'.format(
                                name=f.name, type=f.get_type_str(), default=f.get_default_str())
                    else:
                        yield '{name} ({type})[{default}]: {name} value\n'.format(
                                name=f.name, type=f.get_type_str(), default=f.get_default_str())

        return "\n" \
               "    Data object {}.\n" \
               "\n" \
               "    Args:\n" \
               "        {}\n".format(cls.__name__, '        '.join(args()))

    @staticmethod
    def make_init_signature(cls):
        self_param = Parameter(name='self', kind=Parameter.POSITIONAL_ONLY)  # First param must be self
        return Signature(parameters=[self_param] + [f.as_parameter() for f in cls.__fields__.values()])

    # @classmethod
    # def make_init(cls, new_cls):
    #     def args():
    #         for f in new_cls.__fields__.values():
    #             if f.init:
    #                 if f.is_required():
    #                     yield '{name}: {type}'.format(name=f.name, type=f.get_type_str())
    #                 else:
    #                     yield '{name}: {type} = MISSING'.format(name=f.name, type=f.get_type_str())
    #
    #     def init_kwargs():
    #         for f in new_cls.__fields__.values():
    #             if f.init:
    #                 yield '{name}={name}'.format(name=f.name)
    #
    #     txt = '\n'.join(('def __init__(self, {}, **kwargs):'.format(', '.join(args())),
    #                      '    """{}"""'.format(cls.make_init_docs(new_cls)),
    #                      '    dataclass_cls.init_func(self, {}, **kwargs)'.format(', '.join(init_kwargs())),
    #                      '    ',
    #                      '    if hasattr(self, "__post_init__"):',
    #                      '        self.__post_init__()', ''))
    #
    #     if new_cls.__module__ in sys.modules:
    #         glbls = sys.modules[new_cls.__module__].__dict__
    #     else:
    #         glbls = {}
    #
    #     ns = {}
    #     glbls['dataclass_cls'] = DataclassMeta
    #     glbls['MISSING'] = MISSING
    #     exec(txt, glbls, ns)
    #     init = ns['__init__']
    #     init.__is_dataclass__ = True
    #     return init

    @staticmethod
    def init_func(self, *args, **kwargs):
        # Set positional arguments
        len_args = len(args)
        for i, f in enumerate(self.__fields__.values()):
            if f.init and i < len_args and args[i] != MISSING:  # Positional arguments are prioritized
                setattr(self, f.name, args[i])
            elif f.init and f.name in kwargs:
                if kwargs[f.name] == MISSING:
                    kwargs[f.name] = f.get_default_value(self)
                setattr(self, f.name, kwargs.pop(f.name))
            else:
                setattr(self, f.name, f.get_default_value(self))

        # Check left over keyword arguments:
        if not self.__dataclass_properties__.get('allow_extra_init', True) and len(kwargs) > 0:
            raise TypeError("__init__() got unexpected keyword argument '{}'".format(list(kwargs)[0]))

        # Run post init function
        post_init = getattr(self, '__post_init__', None)
        if callable(post_init):
            post_init()

        # Freeze the object from changing
        if getattr(self, '__is_frozen__', False):
            def setter_func(self, name, value):
                raise TypeError('Cannot set attr on frozen object!')
            self.__setattr__ = setter_func.__get__(self, type(self))

    @staticmethod
    def asdict(self):
        return {f.name: getattr(self, f.name, MISSING) for f in self.__fields__.values()
                if f.dict and f.skip_dict != getattr(self, f.name, MISSING)}

    @staticmethod
    def hash_func(self):
        tup = tuple('{}={}'.format(f.name, getattr(self, f.name, MISSING)) for f in self.__fields__.values()
                    if f.hash or (f.hash is None and f.compare))
        return hash(tup)

    @staticmethod
    def compare_func(self, other):
        if not hasattr(other, '__fields__'):
            return False
        me = {f.name: getattr(self, f.name, MISSING) for f in self.__fields__.values() if f.compare}
        other = {f.name: getattr(other, f.name, MISSING) for f in other.__fields__.values() if f.compare}
        return me == other

    @staticmethod
    def repr_func(self):
        args = ['{}={}'.format(f.name, getattr(self, f.name, 'MISSING'))
                for f in self.__fields__.values()
                if f.repr and f.skip_repr != getattr(self, f.name, MISSING)]
        return '{}({})'.format(self.__class__.__name__, ', '.join(args))

    @staticmethod
    def getstate_func(self):
        return self.dict()

    @staticmethod
    def setstate_func(self, state):
        # Copy may improperly set a default skip_dict values to the field object and not the default.
        # The set defaults below fixes issue with copy.copy on field(skip_dict=default_value)
        for f in self.__fields__.values():
            if not f.is_required():
                # Only set fields that have a default
                setattr(self, f.name, f.get_default_value(self))

        # Set the given state values
        for k, v in state.items():
            setattr(self, k, v)


class DataClass(metaclass=DataclassMeta):
    def __init__(self, *args, **kwargs):  # Defined for IDE argument highlighting
        DataclassMeta.init_func(self, *args, **kwargs)

    __init__.__is_dataclass__ = True


dataclass = DataclassMeta.dataclass


class Message(DataClass):
    pass



import datetime
from serial_json.interface import register


__all__ = ['TIME_FORMATS', 'DATE_FORMATS', 'DATETIME_FORMATS', 'make_time', 'make_date', 'make_datetime']


TIME_FORMATS = [
    '%I:%M:%S %p',     # '02:24:55 PM'
    '%I:%M:%S.%f %p',  # '02:24:55.000200 PM'
    '%I:%M %p',        # '02:24 PM'
    '%H:%M:%S',        # '14:24:55'
    '%H:%M:%S.%f',     # '14:24:55.000200'
    '%H:%M',           # '14:24'
    ]

DATE_FORMATS = [
    '%Y-%m-%d', '%m/%d/%Y',   # '2019-04-17', '04/17/2019'
    '%b %d %Y', '%b %d, %Y',  # 'Apr 17 2019', 'Apr 17, 2019'
    '%d %b %Y', '%d %b, %Y',  # '17 Apr 2019', '17 Apr, 2019'
    '%B %d %Y', '%B %d, %Y',  # 'April 17 2019', 'April 17, 2019'
    '%d %B %Y', '%d %B, %Y',  # '17 April 2019', '17 April, 2019'
    ]

DATETIME_FORMATS = [d + ' ' + t for t in TIME_FORMATS for d in DATE_FORMATS] + DATE_FORMATS + TIME_FORMATS


def make_time(time_string, formats=None):
    """Make the time object from the given time string.

    Args:
        time_string (str): Time string '04:00 PM' ...
        formats (list): List of acceptable time string formats.

    Returns:
        t (datetime.time): Time object or None.
    """
    if isinstance(time_string, datetime.time):
        return time_string

    if formats is None:
        formats = DATETIME_FORMATS

    for fmt in formats:
        try:
            dt = datetime.datetime.strptime(time_string, fmt)
            return dt.time()
        except (TypeError, ValueError, Exception):
            pass

    raise ValueError('Invalid time format {}. Allowed formats are {}'.format(repr(time_string), repr(formats)))


def make_date(date_string, formats=None):
    """Make the date object from the given time string.

    Args:
        date_string (str): Date string 'mm/dd/yyyy' ...
        formats (list): List of acceptable time string formats.

    Returns:
        d (datetime.date): Date object or None.
    """
    if isinstance(date_string, datetime.date):
        return date_string

    if formats is None:
        formats = DATETIME_FORMATS

    for fmt in formats:
        try:
            dt = datetime.datetime.strptime(date_string, fmt)
            return dt.date()
        except (TypeError, ValueError, Exception):
            pass

    try:  # Try ISO format
        return datetime.datetime.fromisoformat(date_string)
    except:
        pass

    raise ValueError('Invalid date format {}. Allowed formats are {}'.format(repr(date_string), repr(formats)))


def make_datetime(date_string, formats=None):
    """Make the datetime from the given date time string.

    Args:
        date_string (str): Datetime string '04:00 PM' ...
        formats (list): List of acceptable datetime string formats.

    Returns:
        dt (datetime.datetime): Datetime object or None.
    """
    if isinstance(date_string, datetime.datetime):
        return date_string

    if formats is None:
        formats = DATETIME_FORMATS

    for fmt in formats:
        try:
            return datetime.datetime.strptime(date_string, fmt)
        except (TypeError, ValueError, Exception):
            pass

    raise ValueError('Invalid datetime format {}. Allowed formats are {}'.format(repr(date_string), repr(formats)))


# ========== Register Datetime compatibility ==========
def date_encode(obj):
    return {'value': obj.strftime(DATE_FORMATS[0])}


def date_decode(obj):
    return make_date(obj['value'])


def time_encode(obj):
    return {'value': obj.strftime(TIME_FORMATS[0])}


def time_decode(obj):
    return make_time(obj['value'])


def datetime_encode(obj):
    return {'value': obj.strftime(DATETIME_FORMATS[0])}


def datetime_decode(obj):
    return make_datetime(obj['value'])


register(datetime.date, date_encode, date_decode)
register(datetime.time, time_encode, time_decode)
register(datetime.datetime, datetime_encode, datetime_decode)



from serial_json.interface import register


def bytes_to_str(bts):
    return bts.decode('latin1')


def bytes_from_str(bts):
    return bts.encode('latin1')


register(bytes, bytes_to_str, bytes_from_str)



name = 'serial_json'
version = '1.1.0'
description = 'Custom json serialization support.'
url = 'https://github.com/justengel/serial_json'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'



from serial_json.interface import Serializer, register, unregister, get_serializer, \
    base_create_object, RegisterMetaclass, \
    dumps, dump, loads, load, default, object_hook

from .dataclasses import MISSING, field, field_property, DataclassMeta, DataClass, dataclass, Message

try:
    import serial_json.bytes_support
except (ImportError, Exception):
    pass

try:
    import serial_json.datetime_support
    TIME_FORMATS = serial_json.datetime_support.TIME_FORMATS
    DATE_FORMATS = serial_json.datetime_support.DATE_FORMATS
    DATETIME_FORMATS = serial_json.datetime_support.DATETIME_FORMATS
    make_time = serial_json.datetime_support.make_time
    make_date = serial_json.datetime_support.make_date
    make_datetime = serial_json.datetime_support.make_datetime
except (ImportError, Exception):
    TIME_FORMATS = []
    DATE_FORMATS = []
    DATETIME_FORMATS = []

    def make_time(*args, **kwargs):
        raise EnvironmentError('Could not properly setup datetime utilities.')

    def make_date(*args, **kwargs):
        raise EnvironmentError('Could not properly setup datetime utilities.')

    def make_datetime(*args, **kwargs):
        raise EnvironmentError('Could not properly setup datetime utilities.')


try:
    from serial_json.numpy_support import np_to_dict, np_from_dict, rec_from_dict
except (ImportError, Exception):
    def np_to_dict(*args, **kwargs):
        raise EnvironmentError('Could not properly setup numpy utilities.')

    def np_from_dict(*args, **kwargs):
        raise EnvironmentError('Could not properly setup numpy utilities.')

    def rec_from_dict(*args, **kwargs):
        raise EnvironmentError('Could not properly setup numpy utilities.')
